from __future__ import absolute_import, division, print_function, unicode_literals

import logging

from credsmash.crypto import seal_secret, open_secret, ALGO_AES_CTR

logger = logging.getLogger(__name__)


def put_secret(
        storage_service, key_service, secret_name,
        plaintext, version=None, compare=True,
        algorithm=ALGO_AES_CTR, **seal_kwargs
):
    sealed = seal_secret(
        key_service,
        plaintext,
        algorithm=algorithm,
        binary_type=storage_service.binary_type,
        **seal_kwargs
    )

    if version is None:
        latest_secret = storage_service.get_one(secret_name)
        version = 1
        if latest_secret:
            version += latest_secret['version']
            if compare:
                latest_plaintext = open_secret(key_service, latest_secret)
                if plaintext == latest_plaintext:
                    logger.debug('secret "%s" is unchanged', secret_name)
                    return latest_secret['version']

    storage_service.put_one(secret_name, version, sealed)
    return version



from __future__ import absolute_import, division, print_function

import codecs
import grp
import json
import logging
import os
import pwd

import click
import jinja2.sandbox

import credsmash.api
from .util import read_many, shell_quote, parse_manifest, detect_format, ItemNotFound
from .cli import main

logger = logging.getLogger(__name__)


class CredsmashProxy(object):
    def __init__(self, key_service, storage_service, key_fmt):
        self._key_service = key_service
        self._storage_service = storage_service
        self._key_fmt = key_fmt
        self._data = {}

    def __getitem__(self, key):
        if key in self._data:
            return self._data[key]

        if isinstance(key, tuple):
            lookup_key = self._key_fmt.format(*key)
        else:
            lookup_key = self._key_fmt.format(key)
        logger.debug('key=%s lookup_key=%s', key, lookup_key)
        res = credsmash.api.get_secret(
            self._storage_service,
            self._key_service,
            key,
        )
        self._data[key] = res
        return res

    def __contains__(self, key):
        try:
            self.__getitem__(key)
            return True
        except ItemNotFound:
            return False


class DictProxy(object):
    def __init__(self, items, key_fmt):
        self._items = items
        self._key_fmt = key_fmt

    def __getitem__(self, key):
        if isinstance(key, tuple):
            lookup_key = self._key_fmt.format(*key)
        else:
            lookup_key = self._key_fmt.format(key)
        return self._items[lookup_key]

    def __contains__(self, key):
        try:
            self.__getitem__(key)
            return True
        except KeyError:
            return False


@main.command('render-template')
@click.argument('template', type=click.File(mode='r', encoding='utf-8'))
@click.argument('destination', type=click.File(mode='w', encoding='utf-8'))
@click.option('--obj-name', default='secrets',
              help='The variable/object name provided to the template')
@click.option('--key-fmt', default='{0}',
              help='Re-use templates by tweaking which variable it maps to- '
                   'eg, "dev.{0}" converts {{secrets.potato}} to the secret "dev.potato"')
@click.option('--template-vars', type=click.File(mode='rb'))
@click.option('--template-vars-format', default=None)
@click.option('--secrets-file', type=click.File(mode='rb'),
              help="Source from a local file instead of credential store "
                   "(useful for caching/testing)")
@click.option('--secrets-file-format', default=None)
@click.pass_context
def cmd_render_template(
        ctx, template, destination,
        obj_name='secrets', key_fmt='{0}',
        template_vars=None, template_vars_format=None,
        secrets_file=None, secrets_file_format=None
):
    """
    Render a configuration template....
    """
    if secrets_file:
        if not secrets_file_format:
            secrets_file_format = detect_format(secrets_file, 'json')
        local_secrets = read_many(secrets_file, secrets_file_format)
        secrets = DictProxy(local_secrets, key_fmt)
    else:
        secrets = CredsmashProxy(
            ctx.obj.key_service,
            ctx.obj.storage_service,
            key_fmt,
        )

    render_args = {}
    if template_vars:
        if not template_vars_format:
            template_vars_format = detect_format(template_vars, 'json')
        render_args = read_many(template_vars, template_vars_format)
    if obj_name in render_args:
        logger.warning('Overwrote %r from template vars with secrets var.', obj_name)
    render_args[obj_name] = secrets

    env = _make_env()
    output = env.from_string(template.read()).render(render_args)
    destination.write(output)


@main.command('render-templates')
@click.argument('manifest', type=click.File(mode='r', encoding='utf-8'))
@click.option('--manifest-format', default=None)
@click.option('--obj-name', default='secrets',
              help='The variable/object name provided to the template')
@click.option('--key-fmt', default='{0}',
              help='Re-use templates by tweaking which variable it maps to- '
                   'eg, "dev.{0}" converts {{secrets.potato}} to the secret "dev.potato"')
@click.option('--template-vars', type=click.File(mode='rb'))
@click.option('--template-vars-format', default=None)
@click.option('--secrets-file', type=click.File(mode='rb'),
              help="Source from a local file instead of credential store "
                   "(useful for caching/testing)")
@click.option('--secrets-file-format', default=None)
@click.pass_context
def cmd_render_template(
        ctx, manifest, manifest_format=None,
        obj_name='secrets', key_fmt='{0}',
        template_vars=None, template_vars_format=None,
        secrets_file=None, secrets_file_format=None
):
    """
    Render multiple configuration templates - reads from a manifest file.
    """
    if secrets_file:
        if not secrets_file_format:
            secrets_file_format = detect_format(secrets_file, 'json')
        local_secrets = read_many(secrets_file, secrets_file_format)
        secrets = DictProxy(local_secrets, key_fmt)
    else:
        secrets = CredsmashProxy(
            ctx.obj.key_service,
            ctx.obj.storage_service,
            key_fmt,
        )

    render_args = {}
    if template_vars:
        if not template_vars_format:
            template_vars_format = detect_format(template_vars, 'json')
        render_args = read_many(template_vars, template_vars_format)
    if obj_name in render_args:
        logger.warning('Overwrote %r from template vars with secrets var.', obj_name)
    render_args[obj_name] = secrets

    env = _make_env()
    if not manifest_format:
        manifest_format = detect_format(manifest, 'json')
    for entry in parse_manifest(manifest, manifest_format):
        if 'source' in entry:
            with codecs.open(entry['source'], 'r', encoding='utf-8') as template:
                output = env.from_string(template.read()).render(render_args)
            # Only open the file after rendering the template
            #  as we truncate the file when opening.
            with codecs.open(entry['destination'], 'w', encoding='utf-8') as destination:
                destination.write(output)
            logger.info('Rendered template="%s" destination="%s"', entry['source'], entry['destination'])
        elif 'secret' in entry:
            output = secrets[entry['secret']]
            with open(entry['destination'], 'wb') as destination:
                destination.write(output)
            logger.info('Wrote secret="%s" destination="%s"', entry['secret'], entry['destination'])
        else:
            raise RuntimeError('Manifest entry must contain a secret or source')

        if 'mode' in entry:
            os.chmod(
                entry['destination'],
                entry['mode']
            )

        if 'owner' in entry and 'group' in entry:
            os.chown(
                entry['destination'],
                pwd.getpwnam(entry['owner']).pw_uid,
                grp.getgrnam(entry['group']).gr_gid
            )


def _make_env():
    env = jinja2.sandbox.SandboxedEnvironment(
        trim_blocks=True,
        lstrip_blocks=True,
        autoescape=False,
    )
    env.filters['sh'] = shell_quote
    env.filters['jsonify'] = json.dumps
    return env



import codecs
import sys
from os.path import dirname, join

from setuptools import setup, find_packages

needs_pytest = {'pytest', 'test', 'ptr'}.intersection(sys.argv)
pytest_runner = ['pytest-runner'] if needs_pytest else []

needs_setupext_pip = {'requirements'}.intersection(sys.argv)
setupext_pip = ['setupext-pip~=1.0.5'] if needs_setupext_pip else []

here = dirname(__file__)


def read(*parts):
    return codecs.open(join(here, *parts), 'r').read()


def find_version(*file_paths):
    version = read(*file_paths).strip()
    if version == '':
        raise RuntimeError('No version found')
    return version


def read_markdown(*file_paths):
    try:
        import pandoc.core
        doc = pandoc.core.Document()
        doc.markdown = read(*file_paths)
        return doc.rst
    except ImportError:
        return ''


setup(
    name='credsmash',
    version=find_version('credsmash', 'VERSION'),

    maintainer="Nathan Muir",
    maintainer_email="ndmuir@gmail.com",

    url='https://github.com/3stack-software/credsmash',

    license='Apache2',
    description='A utility for managing secrets in the cloud using AWS KMS and DynamoDB',
    long_description=read_markdown('README.md'),

    packages=find_packages(exclude=('tests',)),

    package_data={
        'credsmash': ['VERSION']
    },

    setup_requires=[] + pytest_runner + setupext_pip,
    install_requires=[
        'cryptography~=1.5',
        'boto3>=1.1.1',
        'click>=6.6',
        'six'
    ],
    tests_require=[
        'pytest',
    ],
    extras_require={
        'yaml': ['PyYAML>=3.10'],
        'templates': ['jinja2'],
        'documentation': ['pyandoc'],
    },
    entry_points={
        'console_scripts': [
            'credsmash = credsmash.cli:main'
        ],
        'credsmash.key_service': [
            'kms = credsmash.kms_key_service:KmsKeyService',
        ],
        'credsmash.storage_service': [
            'dynamodb = credsmash.dynamodb_storage_service:DynamoDbStorageService',
        ],
        'credsmash.cli': [
            'templates = credsmash.templates',
            'dynamodb = credsmash.cli_dynamodb'
        ]
    },

    classifiers=[
        'Intended Audience :: Developers',
        'Intended Audience :: System Administrators',
        'License :: OSI Approved :: Apache Software License',
        'Programming Language :: Python :: 2',
        'Programming Language :: Python :: 2.7',
    ],
)



from __future__ import absolute_import, division, print_function, unicode_literals

import csv
import json
import logging
import re
import os

import six
from six.moves import configparser

try:
    import yaml
    HAS_YAML = True
except ImportError:
    HAS_YAML = False

logger = logging.getLogger(__name__)


class ItemNotFound(Exception):
    pass


def detect_format(fo, default_format):
    fname = getattr(fo, 'name', None)
    if fname:
        ext = os.path.splitext(fname)[-1].lower()
        if ext == '.json':
            return 'json'
        if ext == '.yml' or ext == '.yaml':
            return 'yaml'
        if ext == '.csv':
            return 'csv'
    return default_format


def write_one(secret_name, secret_value, destination, format):
    if format == 'raw':
        destination.write(secret_value)
    else:
        write_many({secret_name: secret_value}, destination, format)


def write_many(secrets, destination, format):
    if format == 'csv':
        # XX - current csv writer doesn't support unicode, so force to ascii
        secrets = dict(encode_strings(secrets, encoding='ascii'))
        csvwriter = csv.writer(destination)
        for secret_name, secret_value in secrets.items():
            csvwriter.writerow([secret_name, secret_value])
        return

    # Both JSON/YAML support unicode, so this should be fine
    secrets = dict(encode_strings(secrets, encoding='utf-8'))
    if format == 'json':
        json.dump(secrets, destination, sort_keys=True, indent=4, separators=(',', ': '))
        return

    if format == 'yaml':
        if not HAS_YAML:
            raise RuntimeError('YAML Module not loaded. Please install with `pip install credsmash[yaml]`')
        # Using safe_dump prevents it adding all the "!!python/unicode" annotations
        yaml.safe_dump(
            secrets, destination, default_flow_style=False,
            encoding='utf-8', allow_unicode=True
        )
        return

    raise RuntimeError('Unsupported format: %s' % format)


def encode_strings(secrets, encoding):
    for secret_name, secret_value in secrets.items():
        if isinstance(secret_value, six.binary_type):
            try:
                secret_value = secret_value.decode(encoding)
            except UnicodeDecodeError:
                logger.warning('Could not decode %s as utf-8.', secret_name)
                # Skip this secret
                continue
        yield secret_name, secret_value


def read_one(secret_name, source, format):
    if format == 'raw':
        return source.read()

    secrets = read_many(source, format)
    return secrets[secret_name]


def read_many(source, format):
    if format == 'csv':
        csvreader = csv.DictReader(source, ['name', 'value'])
        return {
            entry['name']: entry['value']
            for entry in csvreader
        }

    if format == 'json':
        secrets = json.load(source)
    elif format == 'yaml':
        if not HAS_YAML:
            raise RuntimeError('YAML Module not loaded. Please install with `pip install credsmash[yaml]`')
        secrets = yaml.load(source)
    else:
        raise RuntimeError('Unsupported format: %s' % format)

    if not isinstance(secrets, dict):
        raise RuntimeError('Unsupported type: %s', type(secrets))
    for secret_name, secret_value in secrets.items():
        if not isinstance(secret_value, (str, unicode)):
            raise RuntimeError('Unsupported type: %s=%s', secret_name, type(secret_value))
    return secrets


def parse_manifest(source, format):
    if format == 'json':
        return json.load(source)
    if format == 'yaml':
        if not HAS_YAML:
            raise RuntimeError('YAML Module not loaded. Please install with `pip install credsmash[yaml]`')
        return yaml.load(source)
    raise RuntimeError('Unsupported format: %s' % format)


def parse_config(fp):
    config = {}
    cp = configparser.RawConfigParser()
    cp.readfp(fp)
    for section in cp.sections():
        config[section] = {}
        for option in cp.options(section):
            config_value = cp.get(section, option)
            if config_value.startswith("\n"):
                config_value = _parse_nested(config_value)
            config[section][option] = config_value
    return config


def _parse_nested(config_value):
    # Given a value like this:
    # \n
    # foo = bar
    # bar = baz
    # We need to parse this into
    # {'foo': 'bar', 'bar': 'baz}
    parsed = {}
    for line in config_value.splitlines():
        line = line.strip()
        if not line:
            continue
        # The caller will catch ValueError
        # and raise an appropriate error
        # if this fails.
        key, value = line.split('=', 1)
        parsed[key.strip()] = value.strip()
    return parsed


def set_stream_logger(name='credsmash', level=logging.DEBUG, format_string=None):
    if format_string is None:
        format_string = "%(asctime)s %(name)s [%(levelname)s] %(message)s"

    logger = logging.getLogger(name)
    logger.setLevel(level)
    handler = logging.StreamHandler()
    handler.setLevel(level)
    formatter = logging.Formatter(format_string)
    handler.setFormatter(formatter)
    logger.addHandler(handler)


# Set up logging to ``/dev/null`` like a library is supposed to.
# http://docs.python.org/3.3/howto/logging.html#configuring-logging-for-a-library
class NullHandler(logging.Handler):
    def emit(self, record):
        pass


logging.getLogger('credsmash').addHandler(NullHandler())


_find_unsafe = re.compile(r'[a-zA-Z0-9_^@%+=:,./-]').search


def shell_quote(s):
    """Return a shell-escaped version of the string *s*."""
    if not s:
        return "''"

    if _find_unsafe(s) is None:
        return s

    # use single quotes, and put single quotes into double quotes
    # the string $'b is then quoted as '$'"'"'b'

    return "'" + s.replace("'", "'\"'\"'") + "'"




from __future__ import absolute_import, division, print_function, unicode_literals

import botocore.exceptions


class KmsKeyService(object):
    def __init__(self, session, key_id, encryption_context=None):
        self.kms = session.client('kms')
        self.key_id = key_id
        if not encryption_context:
            encryption_context = {}
        self.encryption_context = encryption_context

    def generate_key_data(self, number_of_bytes):
        try:
            kms_response = self.kms.generate_data_key(
                KeyId=self.key_id, EncryptionContext=self.encryption_context, NumberOfBytes=int(number_of_bytes)
            )
        except:
            raise KmsError("Could not generate key using KMS key %s" % self.key_id)
        return kms_response['Plaintext'], kms_response['CiphertextBlob']

    def decrypt(self, encoded_key):
        try:
            kms_response = self.kms.decrypt(
                CiphertextBlob=encoded_key,
                EncryptionContext=self.encryption_context
            )
        except botocore.exceptions.ClientError as e:
            if e.response["Error"]["Code"] == "InvalidCiphertextException":
                if self.encryption_context is None:
                    msg = ("Could not decrypt hmac key with KMS. The credential may "
                           "require that an encryption context be provided to decrypt "
                           "it.")
                else:
                    msg = ("Could not decrypt hmac key with KMS. The encryption "
                           "context provided may not match the one used when the "
                           "credential was stored.")
            else:
                msg = "Decryption error %s" % e
            raise KmsError(msg)
        return kms_response['Plaintext']

    def __repr__(self):
        return 'KmsKeyService(key_id={0},context={1})'.format(self.key_id, self.encryption_context)


class KmsError(Exception):

    def __init__(self, value=""):
        self.value = "KMS ERROR: " + value if value is not "" else "KMS ERROR"

    def __str__(self):
        return self.value
